"""
	shaman/shaman
	-----------------
	Programming Language Detector

	:author: Prev(prevdev@gmail.com)
	:license: MIT
"""

SUPPORTING_LANGUAGES = ('actionscript', 'asp', 'bash', 'c', 'c#', 'css', 'haxe', 'html', 'java', 'javascript', 'jsp', 'objective-c', 'perl', 'php', 'python', 'ruby', 'sql', 'swift', 'visualbasic', 'xml')

import os, json
import re
import math

class Shaman :

	_default_instance = None

	@staticmethod
	def default() :
		""" Get default shaman instance by "data/trained.json"
		"""
		if Shaman._default_instance is not None :
			return Shaman._default_instance

		with open((os.path.dirname(__file__) or '.') + '/data/trained.json') as file :
			tset = json.loads(file.read())

		Shaman._default_instance = Shaman(tset)
		return Shaman._default_instance


	def __init__(self, trained_set) :
		""" Shaman constructor
		"""
		if hasattr(trained_set, 'read') :
			trained_set = json.loads(trained_set.read())

		self.trained_set = trained_set


	def detect(self, code) :
		""" Detect language with code
		"""
		
		keywords = KeywordFetcher.fetch( code )
		probabilities = {}

		for keyword in keywords :
			if keyword not in self.trained_set['keywords'] :
				continue

			data = self.trained_set['keywords'][keyword]
			p_avg = sum(data.values()) / len(data) # Average probability of all languages

			for language, probability in data.items() :
				# By Naïve Bayes Classification
				p = probability / p_avg

				probabilities[ language ] = probabilities.get(language, 0) + math.log(1 + p)



		for pattern, data in self.trained_set['patterns'].items() :
			# print('pattern ("%s")' % pattern)

			matcher = PatternMatcher(pattern)
			p0 = matcher.getratio(code)
			# print(p0)

			for language, p_avg in data.items() :
				if language not in probabilities :
					continue

				p = 1 - abs(p_avg - p0)
				probabilities[ language ] *= p
				
				# print('%s: %s' % (language, p_avg))

			# print('---------')


		# Convert `log` operated probability to percentile
		sum_val = 0
		for language, p in probabilities.items() :
			sum_val += math.pow(math.e / 2, p)

		for language, p in probabilities.items() :
			probabilities[language] = math.pow(math.e / 2, p) / sum_val * 100

		return sorted(probabilities.items(), key=lambda a: a[1], reverse=True)




class KeywordFetcher :
	
	prog = re.compile('([a-zA-Z0-9$*#@_-]+)')

	@staticmethod
	def fetch(code) :
		""" Fetch keywords by Code
		"""
		ret = {}

		code = KeywordFetcher._remove_strings(code)
		result = KeywordFetcher.prog.findall(code)

		for keyword in result :
			if len(keyword) <= 1: continue # Ignore single-length word
			if keyword.isdigit(): continue # Ignore number

			if keyword[0] == '-' or keyword[0] == '*' : keyword = keyword[1:] # Remove first char if string is starting by '-' or '*' (Pointer or Negative numbers)
			if keyword[-1] == '-' or keyword[-1] == '*' : keyword = keyword[0:-1] # Remove last char if string is finished by '-' or '*'

			if len(keyword) <= 1: continue

			ret[ keyword ] = ret.get(keyword, 0) + 1 # `ret[ keyword ] += 1` with initial value

		return ret


	@staticmethod
	def _remove_strings(code) :
		""" Remove strings in code
		"""
		removed_string = ""
		is_string_now = None
		
		for i in range(0, len(code)-1) :
			append_this_turn = False

			if code[i] == "'" and (i == 0 or code[i-1] != '\\') :
				if is_string_now == "'" :
					is_string_now = None

				elif is_string_now == None :
					is_string_now = "'"
					append_this_turn = True

			elif code[i] == '"' and (i == 0 or code[i-1] != '\\') :
				if is_string_now == '"' :
					is_string_now = None

				elif is_string_now == None :
					is_string_now = '"'
					append_this_turn = True


			if is_string_now == None or append_this_turn == True :
				removed_string += code[i]

		return removed_string



class PatternMatcher :

	PATTERNS = [
		r'(<.+>[^<]*<\/.+>)|<\\?.+>', #markup
		r'([a-zA-Z0-9-_]+)\s*:\s*.*\s*;', #css
		r'def\s+([^(]+)\s*\([^)]*\)\s*:', #python
		r'function\s+\([^)]*\)\s*{[^}]*}', #js-style function
		r'var\s+[a-zA-Z0-9_$]+', #js-style var
		r'\$[a-zA-Z0-9_$]+', #sigil style var
		r'\([^)]*\)\s*{[^}]*}', # c-style block
	]

	def __init__(self, pattern) :
		""" Pattern Matcher Constructor
		:param pattern: regex pattern string
		"""
		self.prog = re.compile(pattern)


	def getratio(self, code) :
		""" Get ratio of code and pattern matched
		"""
		if len(code) == 0 : return 0

		code_replaced = self.prog.sub('', code)
		return (len(code) - len(code_replaced)) / len(code)
	



"""
	shaman/tester
	-----------------
	Test shaman
	
	Usage: 
		python tester.py --help
		python tester.py --test <code_bunch.csv> : Test with code bunch (with answer set)
		python tester.py --file <somecode.java> : Test with single file

	:author: Prev(prevdev@gmail.com)
	:license: MIT
"""

from . import shaman
import sys, os
import csv, json
import argparse


aparser = argparse.ArgumentParser()
aparser.add_argument("-b", "--bunch", help="Test with code bunch (with answer set)")
aparser.add_argument("-f", "--file", help="Test with single file")
args = aparser.parse_args()


def run() :
	if args.file:
		test_with_file(args.file)

	elif args.bunch:
		test_with_bunch(args.bunch)

	else:
		print('to see help, COMMAND "shaman-tester --help"')
		sys.exit()



def test_with_file(filename) :
	""" Detect langauge with single files
		Print top 3 languages by calculated probabilities
	"""
	if not os.path.exists(filename) :
		print('File not exists: ' + filename)
		sys.exit(-1)

	with open(filename, 'r') as file :
		code = file.read()

	probabilities = shaman.Shaman.default().detect( code )
	
	for index, (lang, prob) in enumerate(probabilities) :
		if index > 3: break
		
		print("%s: %.2lf%%" % (lang, prob))


def test_with_bunch(filename) :
	""" Test shaman with code bunch and show statistics
	"""

	if not os.path.exists(filename) :
		print('File not exists: ' + filename)
		sys.exit(-1)


	# Read CSV file
	print('Load CSV file')

	csv.field_size_limit(sys.maxsize) # Set CSV limit to sys.maxsize
	filedata = []
	with open(filename) as csvfile :
		reader = csv.reader(csvfile, delimiter=',')
		for row in reader :
			filedata.append(row)


	detector = shaman.Shaman.default()

	correct = 0
	totals = len(filedata)

	results = {}
	print('Start testing')

	for index, (language, code) in enumerate(filedata) :
		print ('Testing %s/%s     ' % (index, len(filedata)), end="\r")

		if language not in shaman.SUPPORTING_LANGUAGES:
			totals -= 1
			continue

		try :
			glang = detector.detect( code )[0][0]
		except IndexError :
			glang = None

		if language not in results :
			results[ language ] = [0, 0, 0]

		if glang == language :
			correct += 1
			results[ language ][0] += 1

		
		results[ language ][1] += 1
		results[ language ][2] = results[ language ][0] / results[ language ][1]


	
	print("------------------------------------------------")
	print("Accuracy: %.2lf%% (Correct: %d / Valid Data: %d)" % (correct/totals*100, correct, totals))
	print("------------------------------------------------")
	
	results = sorted(results.items(), key=lambda x: x[1][0], reverse=True)
	for lang, l in results :
		print("%s: %.2lf%% (%s/%s)" % (lang, l[2] * 100, l[0], l[1]))

run()



"""
	shaman/trainer
	-----------------
	Train shaman by code bunch

	Usage: 
		python trainer.py <code_bunch.csv> <result.json>

	:author: Prev(prevdev@gmail.com)
	:license: MIT
"""

from . import shaman
import sys, os
import csv, json


def run() :
	if len(sys.argv) != 3 :
		# Exception handling on starting program
		print('Usage: "shaman-trainer <code_bunch.csv> <result.json>"')
		sys.exit(-1)


	# Args
	codebunch_file = sys.argv[1]
	result_file = sys.argv[2]


	if not os.path.isfile(codebunch_file) :
		# Exception handling of <code bunch> file
		print('"%s" is not a file' % codebunch_file)
		sys.exit(-1)



	# Read CSV file
	csv.field_size_limit(sys.maxsize) # Set CSV limit to sys.maxsize
	filedata = []

	print('Load CSV file')

	with open(codebunch_file) as csvfile :
		reader = csv.reader(csvfile, delimiter=',')
		for row in reader :
			filedata.append(row)


	# Fetch keyword data
	trained_data = {}
	trained_data['keywords'] = fetch_keywords(filedata)
	trained_data['patterns'] = match_patterns(filedata)

	# Save result
	with open(result_file, 'w') as file :
		file.write( json.dumps(trained_data) )

	print('Trained result is saved at "%s"' % result_file)


def fetch_keywords(codedata) :
	""" Fetch keywords by shaman.KeywordFetcher
		Get average probabilities of keyword and language
	"""

	# Read row in codedata and count keywords in codes with langauge
	tmp = {}
	language_counts = {}

	for index, (language, code) in enumerate(codedata) :
		if language not in shaman.SUPPORTING_LANGUAGES :
			continue

		if language not in tmp :
			tmp[language] = {}
			language_counts[language] = 0

		language_counts[language] += 1

		for keyword in shaman.KeywordFetcher.fetch( code ) :
			# if keyword exists in fetched data, add '1' to keyword data
			tmp[language][keyword] = tmp[language].get(keyword, 0) + 1

		print('Fetch keyword %d/%d    ' % (index, len(codedata)), end='\r')


	# Get dataset indexed by keyword
	ret = {}

	for language in tmp :
		for keyword, count in tmp[ language ].items() :	
			if keyword not in ret :
				ret[ keyword ] = {}

			ret[ keyword ][ language ] = (count / language_counts[ language ]) # Probability

	print('Fetch keyword completed        ')
	return ret



def match_patterns(codedata) :
	""" Match patterns by shaman.PatternMatcher
		Get average ratio of pattern and language
	"""

	ret = {}

	for index1, pattern in enumerate(shaman.PatternMatcher.PATTERNS) :
		print('Matching pattern %d "%s"' % (index1+1, pattern))

		matcher = shaman.PatternMatcher(pattern)
		tmp = {}

		for index2, (language, code) in enumerate(codedata) :
			if language not in shaman.SUPPORTING_LANGUAGES :
				continue

			if len(code) <= 20 or len(code) > 100000 :
				continue

			if language not in tmp :
				tmp[language] = []

			ratio = matcher.getratio(code)
			tmp[language].append(ratio)

			print('Matching patterns %d/%d    ' % (index2, len(codedata)), end='\r')


		ret[pattern] = {}
		for language, data in tmp.items() :
			ret[pattern][language] = sum(tmp[language]) / max(len(tmp[language]), 1)

	print('Matching patterns completed          ')
	return ret

run()



from setuptools import setup


setup(
	name = 'shamanld',
	packages = ['shamanld'],
	version = '0.11',
	description = 'Programming Language Detector',
	license = 'MIT',

	author = 'Youngsoo Lee',
	author_email = 'prevdev@gmail.com',
	
	url = 'https://github.com/Prev/shaman',
	keywords = ['language-detector', 'language', 'detector'],

	install_requires=[
	],


	classifiers=(
	),

	entry_points={
		'console_scripts': [
			'shaman-trainer = shamanld.trainer:run',
			'shaman-tester = shamanld.tester:run',
		],
	},
)